//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2009-2022 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef f_AT_UIDBGCALLSTACK_H
#define f_AT_UIDBGCALLSTACK_H

#include <at/atcore/deviceprinter.h>

#include "console.h"
#include "debugger.h"
#include "uidbgpane.h"

class IVDTextEditor;
class ATUIPrinterGraphicalOutputWindow;

class ATPrinterOutputWindow : public ATUIPaneWindow
							, public IATPrinterOutput
							, public IATPrinterGraphicalOutput
{
public:
	ATPrinterOutputWindow();
	~ATPrinterOutputWindow();

public:
	void *AsInterface(uint32 id) override;
	int AddRef() override { return ATUIPaneWindow::AddRef(); }
	int Release() override { return ATUIPaneWindow::Release(); }

public:
	bool WantUnicode() const override;
	void WriteRaw(const uint8 *raw, size_t len) override;
	void WriteUnicode(const wchar_t *unicode, size_t len) override;

public:
	void BeginGraphicOutput(float pageWidthMM, float pageVBorderMM, float dotSizeMM, float verticalDotPitchMM, bool bit0Top, uint32 numPins) override;
	void EndGraphicOutput() override;
	void FeedPaper(float distanceMM) override;
	void Print(float x, uint32 dots) override;

protected:
	LRESULT WndProc(UINT msg, WPARAM wParam, LPARAM lParam) override;

	bool OnCreate() override;
	void OnDestroy() override;
	void OnSize() override;
	void OnFontsUpdated() override;
	void OnSetFocus() override;

	vdrefptr<IVDTextEditor> mpTextEditor;
	HWND	mhwndTextEditor;

	uint8		mDropNextChar = 0;
	uint32		mLineBufIdx;
	wchar_t		mLineBuf[133];

	vdrefptr<ATUIPrinterGraphicalOutputWindow> mpGraphicWindow;
};

#endif
